function wcuvars = wcsim(mdl,ufixed,optimopt,varargin)
%WCSIM finds worst case uncertainty values in a Simulink model
%   WCUVARS = WCSIM('mdl') optimizes the USS blocks in the Simulink model
%   'mdl' and returns the optimized values in the structure WCUVARS.  The
%   field names and values of WCUVARS are the names of the uncertain 
%   variables and the optimized values. The optimization objective and 
%   constraint functions are specified with RCTObjective and RCTConstraint
%   blocks. The optimization is performed with FMINCON and requires 
%   repeated simulations of the model. Only uncertain UREAL parameters are 
%   varied in the optimization. They are initialized to the values 
%   specified in the Uncertainty Values field of the USS block. All other 
%   uncertain variable types are held fixed during the optimization at the
%   values specified in the Uncertainty Values field.
%
%   WCUVARS = WCSIM('mdl',UFIXED) specifies additional parameters to hold 
%   fixed during the optimization.  UVARS is a cell array of strings that
%   specifies uncertain UREAL parameters to be held fixed. Set UFIXED=[]
%   to optimize over all UREAL parameters.
%
%   WCUVARS = WCSIM('mdl',UFIXED,OPTIMOPT) specifies optimization options
%   to be passed to FMINCON.  See FMINCON help for details. 
%
%   WCUVARS = WCSIM('mdl',UFIXED,OPTIMOPT,SIMOPT) specifies simulation
%   options to be passed to SIM.  All trailing input arguments of wcsim 
%   are passed directly to SIM.  Set SIM help for details.  WCSIM 
%   currently forces the 'SrcWorkspace' and 'DstWorkspace' options to be
%   set to 'base'.
%
% See also sim, fmincon, optimset, optimget, ufind

% Parse inputs
if nargin==1
    ufixed = [];
    optimopt = [];
    simopt = [];
elseif nargin==2
    optimopt = [];
    simopt = [];    
elseif nargin==3
    simopt = [];
else
    % Trailing inputs are simulation options
    simopt = varargin;
end

% Find all WCSIM Objective and Constraint blocks in the model.
%load_system(mdl);
objblks =  find_system(mdl,'Tag','WCSIMObjective');
conblks =  find_system(mdl,'Tag','WCSIMConstraint');

% Find USS blocks
[uvars,pathinfo] = ufind(mdl);
unames = fieldnames(uvars);
Nunc = length(unames);
ublks = unique(pathinfo(:,1));
Nblks = length(ublks);

% Get uncertainty information
ucell = cell(2,Nunc);
uvaryidx = zeros(Nunc,1);
cnt = 0;
for i=1:Nunc
    % Get uncertainty
    unci = getfield(uvars,unames{i});
    
    % Get initial val of uncert (same uncert can appear in multiple blks)
    ucell{1,i} = unames{i};
    pathidx = find( strcmp(unames{i},pathinfo(:,2)) );
    for j = 1:length(pathidx)
        bh = get_param(pathinfo{pathidx(j),1},'Handle');
        uvalue = get(bh,'UValue');
        uvalue = evalin('base',uvalue);    
        if isempty(uvalue)
            initval = get(unci,'NominalValue');
        else
            initval = usubs(unci,uvalue);
        end        
        if j==1
            ucell{2,i} = initval;
        elseif ~isequal(initval,ucell{2,i})
            warning([unames{i} ' has different values specified in ' ...
                'multiple blocks. Initializing to nominal value.']);
            ucell{2,i} = get(unci,'NominalValue');
        end
    end
        
    % Find varying ureals and their ranges 
    if isa(unci,'ureal') && ~any(strcmp(unames{i},ufixed))
        uvaryidx(i) = 1;
        cnt = cnt+1;
        allrange(cnt,:) = pvget(unci,'Range');
    else
        uvaryidx(i) = 0;
    end
end

% Set parameter initial conditions / fixed values
uvaryidx = find(uvaryidx);
UB = allrange(:,2);
LB = allrange(:,1);
X0 = cell2mat(ucell(2,uvaryidx));

% Get handles to objective/constraint functions
Objective = @(parmvec) LOCALObjective(mdl,objblks,ucell,uvaryidx,...
    simopt,parmvec);
Constraint = @(parmvec) LOCALConstraint(conblks);

% Grab original values in the "Uncertainty Value" dialog box. 
originaluvalue = cell(Nblks,1);
for i = 1:Nblks
    bh = get_param(ublks{i},'Handle');
    originaluvalue{i} = get(bh,'UValue');
end

try 
    % Temporarily change the "Uncertainty Value" dialog box 
    for i = 1:Nblks
        bh = get_param(ublks{i},'Handle');
        set(bh,'UValue','RCASTWCSIMOptimizationParameterVector')
    end

    % Run FMINCON to optimize varying UREALs
    A = []; B = []; Aeq = []; Beq = []; 
    BadPert = fmincon(Objective,X0,A,B,Aeq,Beq,LB,UB,Constraint,optimopt);

    % Build up BadPert
    ucell(2,uvaryidx) = num2cell(BadPert);
    wcuvars = struct(ucell{:});

    % Restore USS dialog boxes to their native states
    for i = 1:Nblks
        bh = get_param(ublks{i},'Handle');
        set(bh,'UValue',originaluvalue{i})
    end    
catch ME
    % Restore USS dialog boxes to their native states
    wcuvars = [];
    for i = 1:Nblks
        bh = get_param(ublks{i},'Handle');
        set(bh,'UValue',originaluvalue{i})
    end    
    throw(ME)
end    
    
    
% ---------- Objective Function  
function obj = LOCALObjective(mdl,objblks,ucell,uvaryidx,simopt,P)

% Set workspace and add on to sim options
% XXX -- Need to check if they input an options structure but not time?
if length(simopt)>2
    simopt{2} = simset(simopt{2},'SrcWorkSpace', 'base', ...
        'DstWorkSpace', 'base');
else
    simopt{2} = simset('SrcWorkSpace', 'base','DstWorkSpace', 'base');
end

%Make structure of parameter values
ucell(2,uvaryidx) = num2cell(P);
pstruct = struct(ucell{:});
assignin('base','RCASTWCSIMOptimizationParameterVector',pstruct);

% Simulate Model 
sim(mdl,simopt{:});

% Compute Objective Function 
% (Retrieve data from objective function blocks)
Nobj = length(objblks);
obj = 0;
for i1=1:Nobj
    blkh = get_param(objblks{i1},'Handle');
    VarName = get(blkh,'VarName');
    VarNameVal = evalin('base',VarName);
    MaxMin = get(blkh,'MaxMin');
    if strcmp(MaxMin,'Minimize');
        obj = obj + VarNameVal.objective.value;
    else
        obj = obj - VarNameVal.objective.value;
    end        
end

% ---------- Constraint Function 
% (LOCALObjective runs the sim and this function simply retrieves data
%  from the constraint blocks)
function [C,Ceq] = LOCALConstraint(conblks) 
    
% Initialize Constraint Evaluations
Ncon = length(conblks);
C = zeros(2*Ncon,1);
Ceq = [];  % Currently not used

% Evaluate Constraint Functions
for i1=1:Ncon
    blkh = get_param(conblks{i1},'Handle');
    VarName = get(blkh,'VarName');
    VarNameVal = evalin('base',VarName);        
    con = VarNameVal.constraint.value;
    conub = VarNameVal.constraint.upperbound;
    conlb = VarNameVal.constraint.lowerbound;
    if isfinite(conub)
        C(2*i1-1) = con-conub;
    else
        C(2*i1-1) = 0;  
    end
    if isfinite(conlb)
        C(2*i1) = conlb-con;
    else
        C(2*i1) = 0;  
    end
end        


% Issues:
%   1) A single uncertain parameter can appear in multiple blocks. What
%      if the user specifies different values in different blocks?
%      What should be use for initializing the variable? Should we error?
%      Right now, I just provide a warning and then set to nominal
%   2) We change the Uncertainty Value field of all USS blocks to be the
%      string 'RCASTWCSIMOptimizationParameterVector'.  If the user
%      hits Ctrl+C then the function will error out without restoring the
%      block to its native state; our default string will remain in the 
%      Uncertainty Value field. The Try/Catch I inserted does not fix
%      this issue.
%   3) SIM is called in the LOCALObjective function and we want the
%      LOCALConstraint function to have access to the simulation results.
%      This raises an issue with the SIM destination workspace. Setting
%      the destination workspace to 'current' or 'parent' will create the
%      results in LOCALObjective and FMINCON, respectively. In neither
%      case would the results be accessible to LOCALConstraint.  Thus
%      it seems like the destination workspace must be 'base'.  I guess
%      we could simply copy results needed for LOCALConstraint into
%      'base' or the model workspace.  Similarly, the Source workspace 
%      seems like it must be 'base' unless we want the user to pass all 
%      required data down to LOCALObjective.  Also see Andy's comments
%      below on this issue.
%   4) Need to do some clean up on the RCTObjective and RCTConstraint
%      blocks (and their call-back functions)
%   5) find_system errors out if mdl is not open. Is there a good
%      solution?  I could try to run find_system. If it errors, 
%      run load_system(mdl) to ensure that the model is open.  I would
%      need to close_system(mdl,0) at the end of the file.  This seems
%      like a sloppy solution.
%   6) I do not have the SIM command working for the single output
%      format. (See SIM help)
%   7) Do we need to have RCTObjective and RCTConstraint blocks?  An
%      alternative interface would be to simply use To Workspace blocks.
%      The user could pass in the handle or path to a To Workspace block
%      and the objective function could be evaluated as the value
%      of the To Workspace block signal at the final time step.  L2
%      norms could be computed using integrator blocks.  Linf norms
%      could be computed using max, delay, and switch blocks.   Similarly
%      constraints could be specified by passing a list of block handles/
%      paths to wcsim.    The constraint functions could be evaluated
%      as the value of the To Workspace block at the final time step.
%

%---------------Notes from Andy about the workspaces 
% The user will have parameters in the Simulink Model.  These must
% be read from "the" workspace.  There is no way they can be in this
% workspace, since this function is deeper than the user's calling
% program.  The user will invoke WCSIM from some workspace. WCSIM
% will invoke FMINCON, and FMINCON will call this OBJECTIVE program
% which will call SIM.  'current' refers to OBJECTIVE.  'parent' seems
% to refer to FMINCON, 'parent/parent' would be WCSIM, and 'p/p/p' would
% be where the user called WCSIM.  I don't see how we can get Simulinkfil
% to possibly read anything from that workspace, therefore the only option
% is that the SrcWorkSpace must be BASE.  We will have to use ASSIGNIN to
% get the parameter-value-structure in the BASE workspace, and the user
% **must** use the BASE workspace for any variables they have.  Overall,
% this seems clumsy.  If the user was doing WCSIM within a function of
% their own, then a solution is the following:  All parameters are of the
% form 'SimVar.ParmName', and the function builds up a structure, S, with
% correct values.  Just before the user executes WCSIM, the user executes
% assignin('base','SimVar',S).
